@@ -6,11 +6,12 @@
 open Findlib;;
 
 exception Usage;;
-
+exception Silent_error;;
+  
 type mode =
     M_use | M_query | M_install | M_remove | M_compiler of string | M_dep
   | M_printconf | M_list | M_browser | M_call of (string*string)
-  | M_doc | M_lint
+  | M_doc | M_lint | M_printppx
 ;;
 
 
@@ -24,6 +25,11 @@ and modifier =
   | Plus
 ;;
 
+let sys_error code arg =
+  if arg = "" then
+    Sys_error (Unix.error_message code)
+  else
+    Sys_error (arg ^ ": " ^ Unix.error_message code)
 
 
 let slashify s =
@@ -201,6 +207,11 @@ let arg n =
 ;;
 
 
+let escape_if_needed s =
+  if String.contains s ' ' then "\"" ^ String.escaped s ^ "\"" else s
+;;
+
+
 let use_package prefix pkgnames =
   (* may raise No_such_package *)
   let pdirs =
@@ -344,8 +355,6 @@ type verbosity =
 
 
 let run_command ?filter verbose cmd args =
-  let escape_if_needed s =
-    if String.contains s ' ' then "\"" ^ String.escaped s ^ "\"" else s in
   let printable_cmd =
     cmd ^ " " ^ String.concat " " (List.map escape_if_needed args) in
   ( match verbose with
@@ -768,7 +777,9 @@ let query_package () =
   let recursive = ref false in
   let descendants = ref false in
   let pp = ref false in
-
+  let qe = ref false in
+  let qo = ref false in
+  
   let packages = ref [] in
 
   let append_predicate s =
@@ -814,6 +825,10 @@ let query_package () =
                 "        prints package names";
       "-help-format", Arg.Unit help_format,
                    "     lists the supported formats for -format";
+      "-qe", Arg.Set qe,
+          "              do not print most errors, just set the exit code";
+      "-qo", Arg.Set qo,
+          "              do not print regular output";
     ]
     (fun p -> packages := !packages @ [p])
 "usage: ocamlfind query [ -predicates <p>  | -format <f> |
@@ -824,34 +839,42 @@ let query_package () =
                          -separator <s>   |
                          -descendants     | -recursive  ] package ...";
 
-  let predicates1 =
-    if !pp then
-      "preprocessor" :: "syntax" :: !predicates
-    else
-      !predicates in
-  let packages1 =
-    if !pp then
-      let predicates2 =
-        List.filter (fun p -> p <> "byte" && p <> "native") predicates1 in
-      select_pp_packages predicates2 !packages
-    else
-      !packages in
-  let eff_packages =
-    if !recursive then begin
-      if !descendants then
-	Fl_package_base.package_users predicates1 packages1
+  try
+    let predicates1 =
+      if !pp then
+        "preprocessor" :: "syntax" :: !predicates
       else
-	package_deep_ancestors predicates1 packages1
-    end
-    else
-      packages1
-  in
-  
-  let answers = expand predicates1 eff_packages !format in
-  
-  print_string !prefix;
-  print_string (String.concat !separator answers);
-  print_string !suffix;
+        !predicates in
+    let packages1 =
+      if !pp then
+        let predicates2 =
+          List.filter (fun p -> p <> "byte" && p <> "native") predicates1 in
+        select_pp_packages predicates2 !packages
+      else
+        !packages in
+    let eff_packages =
+      if !recursive then begin
+        if !descendants then
+          Fl_package_base.package_users predicates1 packages1
+        else
+          package_deep_ancestors predicates1 packages1
+      end
+      else
+        packages1
+    in
+
+    let answers = expand predicates1 eff_packages !format in
+
+    if not !qo then (
+      print_string !prefix;
+      print_string (String.concat !separator answers);
+      print_string !suffix;
+    )
+  with
+    ( Findlib.No_such_package _
+    | Failure _
+    | Sys_error _
+    ) when !qe -> raise Silent_error
 ;;
 
 
@@ -1841,14 +1864,27 @@ let find_owned_files pkg dir =
 	 else
 	   file ^ ".owner" in
        (List.mem owner_file files) && (
-	 let f = open_in (Filename.concat dir owner_file) in
-	 try
-	   let line = input_line f in
-	   let is_my_file = (line = pkg) in
-	   close_in f;
-	   is_my_file
-	 with
-	     exc -> close_in f; raise exc
+         try
+           let fd =
+             Unix.openfile (Filename.concat dir owner_file) [Unix.O_RDONLY] 0 in
+           let f =
+             Unix.in_channel_of_descr fd in
+           try
+             let line = input_line f in
+             let is_my_file = (line = pkg) in
+             close_in f;
+             is_my_file
+           with
+             | End_of_file -> close_in f; false
+             | exc -> close_in f; raise exc
+         with
+           | Unix.Unix_error(Unix.ENOENT,_,_) ->
+               (* the owner file might have been removed by a package
+                  removal that is being done in parallel
+                *)
+               false
+           | Unix.Unix_error(code, _, arg) ->
+               raise(sys_error code arg)
        )
     )
     files
@@ -2100,7 +2136,7 @@ let install_package () =
   (* Create the package directory: *)
   install_create_directory !pkgname pkgdir;
 
-  (* Now copy the files into the package directory: *)
+  (* Now copy the files into the package directory (except META): *)
   List.iter
     (fun p ->
        try
@@ -2117,38 +2153,6 @@ let install_package () =
     )
     pkgdir_list;
 
-  (* Now write the META file: *)
-  let write_meta append_directory dir name =
-    (* If there are patches, write the patched META, else copy the file: *)
-    if !patches = [] then
-      copy_file 
-	~rename:(fun _ -> name)
-        ?append:(if append_directory then
-		   Some("\ndirectory=\"" ^ pkgdir ^ 
-			  "\" # auto-added by ocamlfind\n")
-		 else
-		   None)
-	meta_name
-	dir
-    else (
-      let p = Filename.concat dir name in
-      let patched_pkg = patch_pkg pkgdir meta_pkg !patches in
-      let out = open_out p in
-        Fl_metascanner.print out patched_pkg;
-      if append_directory then
-	output_string out ("\ndirectory=\"" ^ pkgdir ^ 
-			     "\" # auto-added by ocamlfind\n");
-      close_out out;
-      prerr_endline ("Installed " ^ p);
-    )
-  in
-  if not !add_files then (
-    if has_metadir then
-      write_meta true !metadir meta_dot_pkg
-    else
-      write_meta false pkgdir "META";
-  );
-
   (* Copy the DLLs into the libexec directory if necessary *)
   if have_libexec then begin
     List.iter
@@ -2191,6 +2195,38 @@ let install_package () =
       prerr_endline("ocamlfind: [WARNING] You have installed DLLs but the directory " ^ dlldir_norm ^ " is not mentioned in ld.conf");
   end;
 
+  (* Finally, write the META file: *)
+  let write_meta append_directory dir name =
+    (* If there are patches, write the patched META, else copy the file: *)
+    if !patches = [] then
+      copy_file 
+	~rename:(fun _ -> name)
+        ?append:(if append_directory then
+		   Some("\ndirectory=\"" ^ pkgdir ^ 
+			  "\" # auto-added by ocamlfind\n")
+		 else
+		   None)
+	meta_name
+	dir
+    else (
+      let p = Filename.concat dir name in
+      let patched_pkg = patch_pkg pkgdir meta_pkg !patches in
+      let out = open_out p in
+        Fl_metascanner.print out patched_pkg;
+      if append_directory then
+	output_string out ("\ndirectory=\"" ^ pkgdir ^ 
+			     "\" # auto-added by ocamlfind\n");
+      close_out out;
+      prerr_endline ("Installed " ^ p);
+    )
+  in
+  if not !add_files then (
+    if has_metadir then
+      write_meta true !metadir meta_dot_pkg
+    else
+      write_meta false pkgdir "META";
+  );
+
   (* Check if there is a postinstall script: *)
   let postinstall = Filename.concat !destdir "postinstall" in
   if Sys.file_exists postinstall then
@@ -2259,57 +2295,80 @@ let remove_package () =
 	Unix.Unix_error(_,_,_) -> ()    (* ignore, it's only a warning *)
   end;
 
+  (* First remove the META file. If it is already gone, assume that a
+     parallel running removal removed it already.
+   *)
+
   (* If there is a metadir, remove the META file from it: *)
-  if has_metadir then begin
-    let f = Filename.concat !metadir meta_dot_pkg in
-    if Sys.file_exists f then begin
-      Sys.remove f;
-      prerr_endline ("Removed " ^ f);
+  let meta_removal_ok =
+    if has_metadir then (
+      let f = Filename.concat !metadir meta_dot_pkg in
+      try
+        Unix.unlink f;
+        prerr_endline ("Removed " ^ f);
+        true
+      with
+        | Unix.Unix_error(Unix.ENOENT,_,_) ->
+            prerr_endline ("ocamlfind: [WARNING] No such file: " ^ f);
+            false
+        | Unix.Unix_error(code, _, arg) ->
+            raise(sys_error code arg)
+    ) else
+      let f = Filename.concat pkgdir "META" in
+      try
+        Unix.unlink f;
+        prerr_endline ("Removed " ^ f);
+        true
+      with
+        | Unix.Unix_error(Unix.ENOENT,_,_) ->
+            prerr_endline ("ocamlfind: [WARNING] No such file: " ^ f);
+            false
+        | Unix.Unix_error(code, _, arg) ->
+            raise(sys_error code arg) in
+
+  if meta_removal_ok then (
+
+    (* Remove files from libexec directory: *)
+    if have_libexec then begin
+      let dll_files = find_owned_files !pkgname dlldir in
+      List.iter
+        (fun file ->
+           let absfile = Filename.concat dlldir file in
+           Sys.remove absfile;
+           prerr_endline ("Removed " ^ absfile)
+        )
+        dll_files
+    end;
+
+    (* Remove the files from the package directory: *)
+    if Sys.file_exists pkgdir then begin
+      let files = Sys.readdir pkgdir in
+      Array.iter (fun f -> Sys.remove (Filename.concat pkgdir f)) files;
+      Unix.rmdir pkgdir;
+      prerr_endline ("Removed " ^ pkgdir)
     end
     else
-      prerr_endline ("ocamlfind: [WARNING] No such file: " ^ f)
-  end;
+      prerr_endline("ocamlfind: [WARNING] No such directory: " ^ pkgdir);
 
-  (* Remove files from libexec directory: *)
-  if have_libexec then begin
-    let dll_files = find_owned_files !pkgname dlldir in
-    List.iter
-      (fun file ->
-	 let absfile = Filename.concat dlldir file in
-	 Sys.remove absfile;
-	 prerr_endline ("Removed " ^ absfile)
-      )
-      dll_files
-  end;
+    (* Modify ld.conf *)
+    if !ldconf <> "ignore" then begin
+      if Sys.file_exists !ldconf then
+        begin
+          let lines = read_ldconf !ldconf in
+          let d = Fl_split.norm_dir pkgdir in
+          let exists = List.exists (fun p -> Fl_split.norm_dir p = d) lines in
+          if exists then begin
+            let lines' = List.filter (fun p -> Fl_split.norm_dir p <> d) lines in
+            write_ldconf !ldconf lines' []
+          end
+        end
+    end;
 
-  (* Remove the files from the package directory: *)
-  if Sys.file_exists pkgdir then begin
-    let files = Sys.readdir pkgdir in
-    Array.iter (fun f -> Sys.remove (Filename.concat pkgdir f)) files;
-    Unix.rmdir pkgdir;
-    prerr_endline ("Removed " ^ pkgdir)
-  end
-  else
-    prerr_endline("ocamlfind: [WARNING] No such directory: " ^ pkgdir);
-
-  (* Modify ld.conf *)
-  if !ldconf <> "ignore" then begin
-    if Sys.file_exists !ldconf then
-      begin
-	let lines = read_ldconf !ldconf in
-	let d = Fl_split.norm_dir pkgdir in
-	let exists = List.exists (fun p -> Fl_split.norm_dir p = d) lines in
-	if exists then begin
-	  let lines' = List.filter (fun p -> Fl_split.norm_dir p <> d) lines in
-	  write_ldconf !ldconf lines' []
-	end
-      end
-  end;
-
-  (* Check if there is a postremove script: *)
-  let postremove = Filename.concat !destdir "postremove" in
-  if Sys.file_exists postremove then
-    run_command Verbose postremove [ slashify !destdir; !pkgname ]
+    (* Check if there is a postremove script: *)
+    let postremove = Filename.concat !destdir "postremove" in
+    if Sys.file_exists postremove then
+      run_command Verbose postremove [ slashify !destdir; !pkgname ]
+  )
 ;;
 
 
@@ -2360,7 +2419,7 @@ let print_configuration() =
       None ->
 	print_endline "Effective configuration:";
 	Printf.printf "Configuration file:\n    %s\n"
-	  (dir Findlib_config.config_file);
+	  (dir (Findlib.config_file()));
 	Printf.printf "Search path:\n";
 	List.iter
 	  (fun p -> Printf.printf "    %s\n" (dir p))
@@ -2427,7 +2486,7 @@ let lint () =
     (fun s -> if Sys.file_exists s
       then Queue.add s meta_files
       else raise(Arg.Bad (Printf.sprintf "%s: file doesn't exists" s)))
-    "usage: ocamlfind ocamldoc <options> <files>...";
+    "usage: ocamlfind lint <options> <files>...";
 
   let error =
     Queue.fold (fun error file ->
@@ -2439,6 +2498,34 @@ let lint () =
 ;;
 
 
+(** print ppx options *)
+let print_ppx () =
+
+  let packages = ref [] in
+  let predicates = ref [] in
+  let ppx_opts = ref [] in
+
+  let add_pred =
+    Arg.String (fun s -> predicates := !predicates @ (Fl_split.in_words s)) in
+  let add_ppx_opt =
+    Arg.String (fun s -> ppx_opts := !ppx_opts @ [s]) in
+
+  parse_args
+    [ "-predicates", add_pred,
+        " specifies comma-separated list of assumed predicates";
+      "-ppxopt", add_ppx_opt,
+        "<pkg>,<opts>  Append options <opts> to ppx invocation for package <pkg>";
+    ]
+    (fun p -> packages := !packages @ [p])
+"usage: ocamlfind printppx [options] package ...";
+
+  let ppx_commands =
+    process_ppx_spec !predicates !packages !ppx_opts
+  in
+  print_endline (String.concat " " (List.map escape_if_needed ppx_commands))
+;;
+
+
 let rec select_mode () =
   let k = !Arg.current in
   let m_string = try arg (k+1) with Not_found -> raise Usage in
@@ -2460,6 +2547,7 @@ let rec select_mode () =
     | ("printconf"|"-printconf")           -> incr Arg.current; M_printconf
     | ("list"|"-list")                     -> incr Arg.current; M_list
     | ("lint"|"-lint")                     -> incr Arg.current; M_lint
+    | ("printppx"|"-printppx")             -> incr Arg.current; M_printppx
     | "-toolchain" ->
 	let t = try arg (k+2) with Not_found -> raise Usage in
 	Findlib.init ~toolchain:t ();
@@ -2504,6 +2592,7 @@ let main() =
     | M_call(pkg,cmd)  -> ocamlcall pkg cmd
     | M_compiler which -> ocamlc which ()
     | M_lint           -> lint()
+    | M_printppx       -> print_ppx()
   with
     Usage ->
       prerr_endline "Usage: ocamlfind query        [-help | other options] <package_name> ...";
@@ -2520,6 +2609,7 @@ let main() =
       prerr_endline "   or: ocamlfind ocamldoc     [-help | other options] <file> ...";
       prerr_endline "   or: ocamlfind install      [-help | other options] <package_name> <file> ...";
       prerr_endline "   or: ocamlfind remove       [-help | other options] <package_name>";
+      prerr_endline "   or: ocamlfind printppx     [-help | other options] <package_name> ...";
       prerr_endline "   or: ocamlfind printconf    [-help] [variable]";
       prerr_endline "   or: ocamlfind lint         [-help] <file>";
       prerr_endline "   or: ocamlfind list";
@@ -2542,6 +2632,8 @@ let main() =
   | Findlib.Package_loop pkg ->
       prerr_endline ("ocamlfind: Package `" ^ pkg ^ "' requires itself");
       exit 2
+  | Silent_error ->
+      exit 2
 ;;
 
 
